from __future__ import unicode_literals

from ..data.picklemixin import PickleMixIn
from urllib.parse import quote_plus

import os


class BaseSQLConfig(object):
    pass


class SQLConfig(BaseSQLConfig, PickleMixIn):
    """
    SQL Configuration for generating/storing SQL connection strings. Only supports Alchemy & PYODBC
    """

    config_type = None
    conn_types = ['alchemy', 'pyodbc']
    config_types = ['sql', 'accdb', 'dsn', 'conn_str']

    def __init__(self, conn_type=None, conn_str=None, server=None, database=None, credentials=None, accdb_fp=None,
                 dsn_name=None):
        """
        Types of possible SQL connections:
            * conn_type & conn_str
            * server & database (credentials is optional)
            * accdb_fp
            * dsn_name

        :param conn_type: (Optional) [alchemy, pyodbc]
        :param conn_str: (Optional) Raw SQL connection string
        :param server: (Optional) Server in string format
        :param database: (Optional) Database in string format
        :param credentials: (Optional) Credentials instance class
        :param accdb_fp: (Optional) Access database filepath
        :param dsn_name: (Optional) DSN name
        """

        self.conn_type = None
        self.server = None
        self.database = None
        self.credentials = None
        self.accdb_fp = None
        self.dsn_name = None
        self.conn_str = None

        if conn_type and conn_str:
            self.conn_str_config(conn_type, conn_str)
        elif server and database:
            self.sql_config(server, database, credentials)
        elif accdb_fp:
            self.accdb_config(accdb_fp)
        elif dsn_name:
            self.dsn_config(dsn_name)

    def sql_config(self, server, database, credentials=None):
        """
        Store elements for SQL connection string generation

        :param server: Server in string format
        :param database: Database in string format
        :param credentials: (Optional) Credentials instance class
        """

        from ..credentials import Credentials

        if not isinstance(server, str):
            raise ValueError("'server' must be a String")
        if not isinstance(database, str):
            raise ValueError("'database' must be a String")
        if not isinstance(credentials, (Credentials, type(None))):
            raise ValueError("'credentials' %r must be a Credentials instance" % credentials)

        self.__slots__ = ("conn_type", "server", "database")
        self.config_type = 'sql'
        self.conn_type = 'alchemy'
        self.server = server
        self.database = database
        self.credentials = credentials
        self.conn_str = None
        self.accdb_fp = None
        self.dsn_name = None

    def accdb_config(self, accdb_fp):
        """
        Store elements for SQL connection string generation

        :param accdb_fp: Access database filepath
        """

        if not isinstance(accdb_fp, str):
            raise ValueError("'accdb_fp' must be a String")
        if not os.path.exists(accdb_fp):
            raise ValueError("'accdb_fp' (%s) filepath does not exist" % accdb_fp)
        if accdb_fp.find('.') < 1:
            raise ValueError("'accdb_fp' (%s) file does not have an extension" % os.path.basename(accdb_fp))

        ext = os.path.splitext(accdb_fp)[1].lower()

        if ext not in ('.accdb', '.mdb'):
            raise ValueError("'accdb_fp' (%s) extension must be either (accdb, mdb)" %
                             os.path.basename(accdb_fp))

        self.__slots__ = ("conn_type", "accdb_fp")
        self.conn_type = 'pyodbc'
        self.config_type = 'accdb'
        self.accdb_fp = accdb_fp
        self.conn_str = None
        self.server = None
        self.database = None
        self.credentials = None
        self.dsn_name = None

    def dsn_config(self, dsn_name):
        """
        Store elements for SQL connection string generation

        :param dsn_name: DSN name
        """
        if not isinstance(dsn_name, str):
            raise ValueError("'dsn_name' must be a String")

        self.__slots__ = ("conn_type", "dsn_name")
        self.conn_type = 'pyodbc'
        self.config_type = 'dsn'
        self.dsn_name = dsn_name
        self.conn_str = None
        self.server = None
        self.database = None
        self.credentials = None
        self.accdb_fp = None

    def conn_str_config(self, conn_type, conn_str):
        """
        Store elements for SQL connection string generation

        :param conn_type: [alchemy, pyodbc]
        :param conn_str: Raw SQL connection string
        """

        if conn_type not in self.conn_types:
            raise ValueError("'conn_type' must be either (%s)" % ', '.join(self.conn_types))
        if not isinstance(conn_str, str):
            raise ValueError("'conn_str' must be a String")

        self.__slots__ = ("conn_type", "conn_str")
        self.config_type = 'conn_str'
        self.conn_type = conn_type
        self.conn_str = conn_str
        self.server = None
        self.database = None
        self.credentials = None
        self.accdb_fp = None
        self.dsn_name = None

    def gen_conn_str(self):
        """
        Generates a connection string for sql, accdb, and dsn connections

        :return: connection string
        """

        if self.config_type == 'sql':
            if self.credentials:
                quote = quote_plus(
                    'DRIVER={};PORT={};SERVER={};DATABASE={};Uid={},Pwd={};Trusted_Connection=yes;'.format(
                        '{SQL Server Native Client 11.0}', '1433', self.server, self.database,
                        self.credentials.username.decrypt(), self.credentials.password.decrypt())
                )
            else:
                quote = quote_plus(
                    'DRIVER={};PORT={};SERVER={};DATABASE={};Trusted_Connection=yes;'.format(
                        '{SQL Server Native Client 11.0}', '1433', self.server, self.database)
                )
            return '{}+pyodbc:///?odbc_connect={}'.format('mssql', quote)
        elif self.config_type == 'accdb':
            return 'DRIVER={};DBQ={};Exclusive=1'.format('{Microsoft Access Driver (*.mdb, *.accdb)}',
                                                         self.accdb_fp)
        elif self.config_type == 'dsn':
            return 'DSN={};DATABASE=default;Trusted_Connection=Yes;'.format(self.dsn_name)
        elif self.config_type == 'conn_str':
            return self.conn_str
        else:
            return None

    def __eq__(self, other):
        if self.__slots__:
            for k in self.__slots__:
                if getattr(self, k) != getattr(other, k):
                    return False

            return True
        else:
            return False

    def __hash__(self):
        items = list()

        for item in self.__slots__:
            items.append(hash(getattr(self, item)))

        return items

    def __repr__(self):
        items = list()

        for item in self.__slots__:
            items.append(str(getattr(self, item)))

        return self.__class__.__name__ + repr(items)

    def __str__(self):
        items = list()

        for item in self.__slots__:
            items.append(str(getattr(self, item)))

        return ', '.join(items)



from ..filehandler import FileHandler
from ..util import maybe_convert_usecols, get_version, Scalar
from distutils.version import LooseVersion
from pandas import to_datetime, Timestamp
from typing import List
from numpy import nan


class OpenPYXLReader(FileHandler):
    def __init__(self, file_path, flag_read_only=False, sheet_names=None, header=0, names=None, index_col=None,
                 usecols=None, squeeze=False, dtype=None, true_values=None, false_values=None, skiprows=0, nrows=0,
                 na_values=None, verbose=False, parse_dates=False, date_parser=None, thousands=None, comment=None,
                 skipfooter=0, convert_float=True, mangle_dupe_cols=True, **kwargs):
        super().__init__(file_path=file_path, flag_read_only=flag_read_only, sheet_names=sheet_names, header=header,
                         names=names, index_col=index_col, usecols=usecols, squeeze=squeeze, dtype=dtype,
                         true_values=true_values, false_values=false_values, skiprows=skiprows, nrows=nrows,
                         na_values=na_values, verbose=verbose, parse_dates=parse_dates, date_parser=date_parser,
                         thousands=thousands, comment=comment, skipfooter=skipfooter, convert_float=convert_float,
                         mangle_dupe_cols=mangle_dupe_cols, **kwargs)

    def parse(self):
        if self.streams:
            for handler, buffer in self.streams:
                self.__parse(handler=handler, buffer=buffer)
        else:
            return self.__parse()

    def __parse(self, handler=None, buffer=None):
        from openpyxl import load_workbook

        xls_file = load_workbook(self.file_path)
        sheets = dict()

        try:
            if self.flag_read_only and xls_file.read_only:
                raise ReadOnlyError

            for tab in xls_file.sheetnames:
                if not self.sheet_names or tab.lower() in self.sheet_names:
                    xls_ws = xls_file[tab]

                    if handler:
                        self.__openpyxl_sheet(sheet=xls_ws, handler=handler, buffer=buffer)
                    else:
                        sheets[tab] = self.__openpyxl_sheet(sheet=xls_ws, handler=handler, buffer=buffer)
        finally:
            xls_file.close()

        if sheets:
            return sheets

    def __openpyxl_sheet(self, sheet, handler=None, buffer=None):
        import openpyxl

        version = LooseVersion(get_version(openpyxl))
        is_readonly = hasattr(sheet, "reset_dimensions")

        if version >= "3.0.0" and is_readonly:
            sheet.reset_dimensions()

        data: List[List[Scalar]] = []
        row_number = 0
        last_row_with_data = -1
        header = None

        for row in sheet.rows:
            if 0 <= self.skiprows <= row_number and 0 <= self.header <= row_number and (row_number <= self.nrows
                                                                                        or self.nrows == 0):
                converted_row = [self.__convert_cell(cell, self.convert_float) for cell in row]
                data.append(converted_row)

                if not header:
                    header = converted_row

                if not all(cell == "" for cell in converted_row):
                    last_row_with_data = len(data)

                if handler and buffer < len(data):
                    data = self.__read_only_fix(data, version, is_readonly)
                    df = self.convert_data(data)
                    handler(df, row_number - len(df) + 1, row_number)
                    data.clear()
                    data.append(header)
                    last_row_with_data = -1

            row_number += 1

        data = data[: last_row_with_data + 1]
        data = self.__read_only_fix(data, version, is_readonly)
        df = self.convert_data(data)

        if not df.empty and handler:
            row_number -= 1
            handler(df, row_number - len(df) + 1, row_number)
        elif handler is None:
            return df

    @staticmethod
    def __read_only_fix(data, version, is_readonly):
        if version >= "3.0.0" and is_readonly and len(data) > 0:
            max_width = max(len(data_row) for data_row in data)

            if min(len(data_row) for data_row in data) < max_width:
                empty_cell: List[Scalar] = [""]
                data = [
                    data_row + (max_width - len(data_row)) * empty_cell
                    for data_row in data
                ]

        return data

    @staticmethod
    def __convert_cell(cell, convert_float: bool) -> Scalar:
        from openpyxl.cell.cell import TYPE_BOOL, TYPE_ERROR, TYPE_NUMERIC

        if cell.value is None:
            return ""
        elif cell.is_date:
            return cell.value
        elif cell.data_type == TYPE_ERROR:
            return nan
        elif cell.data_type == TYPE_BOOL:
            return bool(cell.value)
        elif cell.data_type == TYPE_NUMERIC:
            if convert_float:
                val = int(cell.value)

                if val == cell.value:
                    return val
            else:
                return float(cell.value)

        return cell.value


class ODFReader(FileHandler):
    __odf_file = None

    def __init__(self, file_path, flag_read_only=False, sheet_names=None, header=0, names=None, index_col=None,
                 usecols=None, squeeze=False, dtype=None, true_values=None, false_values=None, skiprows=0, nrows=0,
                 na_values=None, verbose=False, parse_dates=False, date_parser=None, thousands=None, comment=None,
                 skipfooter=0, convert_float=True, mangle_dupe_cols=True, **kwargs):
        super().__init__(file_path=file_path, flag_read_only=flag_read_only, sheet_names=sheet_names, header=header,
                         names=names, index_col=index_col, usecols=usecols, squeeze=squeeze, dtype=dtype,
                         true_values=true_values, false_values=false_values, skiprows=skiprows, nrows=nrows,
                         na_values=na_values, verbose=verbose, parse_dates=parse_dates, date_parser=date_parser,
                         thousands=thousands, comment=comment, skipfooter=skipfooter, convert_float=convert_float,
                         mangle_dupe_cols=mangle_dupe_cols, **kwargs)

    @property
    def empty_value(self) -> str:
        return ""

    def parse(self):
        if self.streams:
            for handler, buffer in self.streams:
                self.__parse(handler=handler, buffer=buffer)
        else:
            return self.__parse()

    def __parse(self, handler=None, buffer=None):
        from odf.opendocument import load
        from odf.table import Table

        self.__odf_file = load(self.file_path)
        sheets = dict()

        try:
            for xls_ws in self.__odf_file.getElementsByType(Table):
                tab = xls_ws.getAttribute("name")

                if not self.__sheet_names or tab.lower() in self.__sheet_names:
                    if handler:
                        self.__odf_sheet(sheet=xls_ws, handler=handler, buffer=buffer)
                    else:
                        sheets[tab] = self.__odf_sheet(sheet=xls_ws, handler=handler, buffer=buffer)
        finally:
            self.__odf_file.close()

        if sheets:
            return sheets

    def __odf_sheet(self, sheet, handler=None, buffer=None):
        from odf.table import CoveredTableCell, TableCell, TableRow

        covered_cell_name = CoveredTableCell().qname
        table_cell_name = TableCell().qname
        cell_names = {covered_cell_name, table_cell_name}

        sheet_rows = sheet.getElementsByType(TableRow)
        header = None
        empty_rows = 0
        max_row_len = 0
        row_number = 0

        table: List[List[Scalar]] = []

        for sheet_row in sheet_rows:
            if 0 < self.skiprows <= row_number <= self.nrows and 0 < self.header <= row_number:
                sheet_cells = [x for x in sheet_row.childNodes if x.qname in cell_names]
                empty_cells = 0
                table_row: List[Scalar] = []

                for j, sheet_cell in enumerate(sheet_cells):
                    if sheet_cell.qname == table_cell_name:
                        value = self.__get_cell_value(sheet_cell)
                    else:
                        value = self.empty_value

                    column_repeat = self.__get_column_repeat(sheet_cell)

                    # Queue up empty values, writing only if content succeeds them
                    if value == self.empty_value:
                        empty_cells += column_repeat
                    else:
                        table_row.extend([self.empty_value] * empty_cells)
                        empty_cells = 0
                        table_row.extend([value] * column_repeat)

                if max_row_len < len(table_row):
                    max_row_len = len(table_row)

                row_repeat = self.__get_row_repeat(sheet_row)

                if self.__is_empty_row(sheet_row):
                    empty_rows += row_repeat
                else:
                    table.extend([[self.empty_value]] * empty_rows)
                    empty_rows = 0

                    for _ in range(row_repeat):
                        table.append(table_row)

                if not header:
                    header = converted_row

                if handler and buffer < len(table):
                    table = self.__fix_table(table)
                    df = self.convert_data(table)
                    handler(df, row_number - len(df) + 1, row_number)
                    table.clear()
                    table.append(header)

            row_number += 1

        table = self.__fix_table(table)
        df = self.convert_data(table)

        if not df.empty and handler:
            row_number -= 1
            handler(df, row_number - len(df) + 1, row_number)
        elif not handler:
            return df

    def __fix_table(self, table):
        for row in table:
            if len(row) < max_row_len:
                row.extend([self.empty_value] * (max_row_len - len(row)))

        return table

    def __get_cell_value(self, cell) -> Scalar:
        from odf.namespaces import OFFICENS

        if str(cell) == "#N/A":
            return nan

        cell_type = cell.attributes.get((OFFICENS, "value-type"))

        if cell_type == "boolean":
            if str(cell) == "TRUE":
                return True

            return False
        if cell_type is None:
            return self.empty_value
        elif cell_type == "float":
            cell_value = float(cell.attributes.get((OFFICENS, "value")))

            if self.convert_float:
                val = int(cell_value)
                if val == cell_value:
                    return val

            return cell_value
        elif cell_type == "percentage":
            cell_value = cell.attributes.get((OFFICENS, "value"))
            return float(cell_value)
        elif cell_type == "string":
            return self.__get_cell_string_value(cell)
        elif cell_type == "currency":
            cell_value = cell.attributes.get((OFFICENS, "value"))
            return float(cell_value)
        elif cell_type == "date":
            cell_value = cell.attributes.get((OFFICENS, "date-value"))
            return to_datetime(cell_value)
        elif cell_type == "time":
            result = to_datetime(str(cell))
            result = cast(Timestamp, result)
            return result.time()
        else:
            self.__odf_file.close()
            raise ValueError(f"Unrecognized type {cell_type}")

    def _get_cell_string_value(self, cell) -> str:
        from odf.element import Element
        from odf.namespaces import TEXTNS
        from odf.text import S

        text_s = S().qname

        value = []

        for fragment in cell.childNodes:
            if isinstance(fragment, Element):
                if fragment.qname == text_s:
                    spaces = int(fragment.attributes.get((TEXTNS, "c"), 1))
                    value.append(" " * spaces)
                else:
                    value.append(self.__get_cell_string_value(fragment))
            else:
                value.append(str(fragment))
        return "".join(value)

    @staticmethod
    def __get_row_repeat(row) -> int:
        from odf.namespaces import TABLENS

        return int(row.attributes.get((TABLENS, "number-rows-repeated"), 1))

    @staticmethod
    def __get_column_repeat(cell) -> int:
        from odf.namespaces import TABLENS

        return int(cell.attributes.get((TABLENS, "number-columns-repeated"), 1))

    @staticmethod
    def __is_empty_row(row) -> bool:
        for column in row.childNodes:
            if len(column.childNodes) > 0:
                return False

        return True


class XLRDReader(FileHandler):
    __xls_file = None

    def __init__(self, file_path, flag_read_only=False, sheet_names=None, header=0, names=None, index_col=None,
                 usecols=None, squeeze=False, dtype=None, true_values=None, false_values=None, skiprows=0, nrows=0,
                 na_values=None, verbose=False, parse_dates=False, date_parser=None, thousands=None, comment=None,
                 skipfooter=0, convert_float=True, mangle_dupe_cols=True, **kwargs):
        super().__init__(file_path=file_path, flag_read_only=flag_read_only, sheet_names=sheet_names, header=header,
                         names=names, index_col=index_col, usecols=usecols, squeeze=squeeze, dtype=dtype,
                         true_values=true_values, false_values=false_values, skiprows=skiprows, nrows=nrows,
                         na_values=na_values, verbose=verbose, parse_dates=parse_dates, date_parser=date_parser,
                         thousands=thousands, comment=comment, skipfooter=skipfooter, convert_float=convert_float,
                         mangle_dupe_cols=mangle_dupe_cols, **kwargs)

    def parse(self):
        if self.streams:
            for handler, buffer in self.streams:
                self.__parse(handler=handler, buffer=buffer)
        else:
            return self.__parse()

    def __parse(self, handler=None, buffer=None):
        from xlrd import open_workbook

        sheets = dict()
        self.__xls_file = open_workbook(self.file_path)

        try:
            for tab in self.__xls_file.sheet_names():
                if not self.sheet_names or tab.lower() in self.sheet_names:
                    xls_ws = xls_file.sheet_by_name(tab)

                    if handler:
                        self.__xlrd_sheet(sheet=xls_ws, handler=handler, buffer=buffer)
                    else:
                        sheets[tab] = self.__xlrd_sheet(sheet=xls_ws, handler=handler, buffer=buffer)
        finally:
            self.__xls_file.close()

        if sheets:
            return sheets

    def __xlrd_sheet(self, sheet, handler=None, buffer=None):
        data: List[List[Scalar]] = []
        row_number = 0
        header = None

        for i in range(sheet.nrows):
            if 0 < self.skiprows <= row_number <= self.nrows and 0 < self.header <= row_number:
                row = [
                    self.__parse_cell(value, typ)
                    for value, typ in zip(sheet.row_values(i), sheet.row_types(i))
                ]
                data.append(row)

                if not header:
                    header = converted_row

                if handler and len(data) <= buffer:
                    df = self.convert_data(data)
                    handler(df, row_number - len(df) + 1, row_number)
                    data.clear()
                    data.append(header)

            row_number += 1

        df = self.convert_data(data)

        if not df.empty and handler:
            row_number -= 1
            handler(df, row_number - len(df) + 1, row_number)
        elif not handler:
            return df

    def __parse_cell(self, cell_contents, cell_typ):
        from xlrd import (
            XL_CELL_BOOLEAN,
            XL_CELL_DATE,
            XL_CELL_ERROR,
            XL_CELL_NUMBER,
            xldate,
        )

        epoch1904 = self.__xls_file.datemode

        if cell_typ == XL_CELL_DATE:
            try:
                cell_contents = xldate.xldate_as_datetime(cell_contents, epoch1904)
            except OverflowError:
                return cell_contents

            year = (cell_contents.timetuple())[0:3]

            if (not epoch1904 and year == (1899, 12, 31)) or (
                    epoch1904 and year == (1904, 1, 1)
            ):
                cell_contents = time(
                    cell_contents.hour,
                    cell_contents.minute,
                    cell_contents.second,
                    cell_contents.microsecond,
                )
        elif cell_typ == XL_CELL_ERROR:
            cell_contents = np.nan
        elif cell_typ == XL_CELL_BOOLEAN:
            cell_contents = bool(cell_contents)
        elif convert_float and cell_typ == XL_CELL_NUMBER:
            val = int(cell_contents)

            if val == cell_contents:
                cell_contents = val

        return cell_contents


class PYXLSBReader(FileHandler):
    def __init__(self, file_path, flag_read_only=False, sheet_names=None, header=0, names=None, index_col=None,
                 usecols=None, squeeze=False, dtype=None, true_values=None, false_values=None, skiprows=0, nrows=0,
                 na_values=None, verbose=False, parse_dates=False, date_parser=None, thousands=None, comment=None,
                 skipfooter=0, convert_float=True, mangle_dupe_cols=True, **kwargs):
        super().__init__(file_path=file_path, flag_read_only=flag_read_only, sheet_names=sheet_names, header=header,
                         names=names, index_col=index_col, usecols=usecols, squeeze=squeeze, dtype=dtype,
                         true_values=true_values, false_values=false_values, skiprows=skiprows, nrows=nrows,
                         na_values=na_values, verbose=verbose, parse_dates=parse_dates, date_parser=date_parser,
                         thousands=thousands, comment=comment, skipfooter=skipfooter, convert_float=convert_float,
                         mangle_dupe_cols=mangle_dupe_cols, **kwargs)

    def parse(self):
        if self.streams:
            for handler, buffer in self.streams:
                self.__parse(handler=handler, buffer=buffer)
        else:
            return self.__parse()

    def __parse(self, handler=None, buffer=None):
        from pyxlsb import open_workbook

        sheets = dict()
        self.__xls_file = open_workbook(self.file_path)

        try:
            for tab in self.__xls_file.sheets:
                if not self.sheet_names or tab.lower() in self.sheet_names:
                    xls_ws = xls_file.get_sheet(tab)

                    if handler:
                        self.__pyxlsb_sheet(sheet=xls_ws, handler=handler, buffer=buffer)
                    else:
                        sheets[tab] = self.__pyxlsb_sheet(sheet=xls_ws, handler=handler, buffer=buffer)
        finally:
            self.__xls_file.close()

        if sheets:
            return sheets

    def __pyxlsb_sheet(self, sheet, handler, buffer):
        data: List[List[Scalar]] = []
        row_number = 0
        header = None

        for r in sheet.rows(sparse=False):
            if 0 < self.skiprows <= row_number <= self.nrows and 0 < self.header <= row_number:
                row = [self.__convert_cell(c) for c in r]
                data.append(row)

                if not header:
                    header = converted_row

                if handler and buffer < len(data):
                    df = self.convert_data(data)
                    handler(df, row_number - len(df) + 1, row_number)
                    data.clear()
                    data.append(header)

            row_number += 1

        df = self.convert_data(data)

        if not df.empty and handler:
            row_number -= 1
            handler(df, row_number - len(df) + 1, row_number)
        elif not handler:
            return df

    def __convert_cell(self, cell) -> Scalar:
        if cell.v is None:
            return ""
        if isinstance(cell.v, float) and self.convert_float:
            val = int(cell.v)

            if val == cell.v:
                return val
            else:
                return float(cell.v)

        return cell.v



from __future__ import unicode_literals

from independentsoft.msg import Message as Msg
from independentsoft.msg import Recipient, ObjectType, DisplayType, AttachmentMethod, RecipientType, MessageFlag,\
    Attachment, Importance
from exchangelib import FileAttachment, ItemAttachment
from exchangelib import Message as Msg2, Account
from bs4 import BeautifulSoup
from os.path import exists, dirname
from logging import getLogger, WARNING

getLogger("exchangelib").setLevel(WARNING)


class ExchangeToMsg(object):
    def __init__(self, ex_lib_item):
        if not isinstance(ex_lib_item, Msg2):
            ValueError('Exchangelib object is not an item class from Exchangelib library')

        self.ex_lib_item = ex_lib_item
        self.subject = self.ex_lib_item.subject

        if self.ex_lib_item.datetime_received:
            self.received_time = self.ex_lib_item.datetime_received.replace(tzinfo=None)
        elif self.ex_lib_item.datetime_created:
            self.received_time = self.ex_lib_item.datetime_created.replace(tzinfo=None)
        else:
            self.received_time = None

        self.message = Msg()
        self.__msg_attachs = []

        self.__mail_attr()
        self.__rec_proc(self.ex_lib_item.to_recipients, RecipientType.TO)
        self.__rec_proc(self.ex_lib_item.cc_recipients, RecipientType.CC)
        self.__mail_attach()
        self.__attach_mail_objs()

    def __mail_attr(self):
        self.message.object_type = ObjectType.MESSAGE

        if self.ex_lib_item.importance == 'Normal':
            self.message.importance = Importance.NORMAL
        elif self.ex_lib_item.importance == 'High':
            self.message.importance = Importance.HIGH
        elif self.ex_lib_item.importance == 'Low':
            self.message.importance = Importance.LOW
        else:
            self.message.importance = Importance.NONE

        if self.received_time:
            self.message.client_submit_time = self.received_time
            self.message.message_delivery_time = self.received_time

        if self.ex_lib_item.sender:
            self.message.reply_to = self.ex_lib_item.sender.email_address
            self.message.in_reply_to = self.ex_lib_item.sender.email_address
            self.message.sender_address_type = self.ex_lib_item.sender.routing_type
            self.message.sender_name = self.ex_lib_item.sender.name
            self.message.sender_search_key = '{0}:{1}\x00'.format(self.ex_lib_item.sender.routing_type,
                                                                  self.ex_lib_item.sender.email_address).encode()
            self.message.sender_smtp_address = self.ex_lib_item.sender.email_address
            self.message.sender_email_address = self.ex_lib_item.sender.email_address
            self.message.sent_address_type = self.ex_lib_item.sender.routing_type
            self.message.sent_name = self.ex_lib_item.sender.name
            self.message.sent_search_key = '{0}:{1}\x00'.format(self.ex_lib_item.sender.routing_type,
                                                                self.ex_lib_item.sender.email_address).encode()
            self.message.sent_smtp_address = self.ex_lib_item.sender.email_address
            self.message.sent_email_address = self.ex_lib_item.sender.email_address

        if self.ex_lib_item.display_to:
            self.message.display_to = self.ex_lib_item.display_to

        if self.ex_lib_item.display_cc:
            self.message.display_cc = self.ex_lib_item.display_cc

        if self.ex_lib_item.has_attachments:
            self.message.has_attachment = self.ex_lib_item.has_attachments

        self.message.subject = self.subject
        self.message.message_flags.append(MessageFlag.HAS_ATTACHMENT)
        self.message.message_flags.append(MessageFlag.READ)

        if self.ex_lib_item.text_body:
            self.message.body_html_text = self.ex_lib_item.text_body
        elif self.ex_lib_item.body:
            if bool(BeautifulSoup(self.ex_lib_item.body, 'html.parser').find()):
                self.message.body_html_text = self.ex_lib_item.body
            else:
                self.message.body_html_text = self.ex_lib_item.body
        else:
            self.message.body_html_text = None

    def __rec_proc(self, recs, rec_type):
        if recs:
            for rec in recs:
                self.__pack_rec(rec_type, rec.routing_type, rec.name, rec.email_address)

    def __pack_rec(self, rec_type, rout_type, rec_name, rec_email):
        rec = Recipient()
        rec.address_type = rout_type
        rec.display_type = DisplayType.MAIL_USER
        rec.object_type = ObjectType.MAIL_USER
        rec.display_name = rec_name
        rec.email_address = rec_email
        rec.smtp_address = rec_email
        rec.recipient_type = rec_type
        self.message.recipients.append(rec)

    def __mail_attach(self):
        if self.ex_lib_item.attachments:
            for attach in self.ex_lib_item.attachments:
                if isinstance(attach, FileAttachment):
                    with attach.fp as fp:
                        buffer = fp.read()

                    sub_attach = Attachment(buffer=buffer)
                    sub_attach.file_name = attach.name

                    if attach.is_inline:
                        sub_attach.content_id = attach.content_id
                        sub_attach.method = AttachmentMethod.EMBEDDED_MESSAGE

                    self.message.attachments.append(sub_attach)
                elif isinstance(attach, ItemAttachment) and isinstance(attach.item, Msg2):
                    self.__msg_attachs.append(ExchangeToMsg(attach.item))

    def __attach_mail_objs(self):
        if self.__msg_attachs:
            for msg in self.__msg_attachs:
                message, subject, received_time = msg.grab_objs()

                # This attaches the message as an embedded message, but it is missing a sequence somewhere
                '''
                sub_attach = Attachment()
                message.message_flags.append(MessageFlag.HAS_ATTACHMENT)
                message.message_flags.append(MessageFlag.READ)
                message.is_embedded = False
                sub_attach.embedded_message = message
                sub_attach.display_name = subject
                sub_attach.file_name = subject
                sub_attach.method = AttachmentMethod.EMBEDDED_MESSAGE
                sub_attach.flags = AttachmentFlags.NONE
                sub_attach.object_type = ObjectType.ATTACHMENT
                sub_attach.creation_time = datetime.datetime.now()
                sub_attach.last_modification_time = received_time
                '''
                sub_attach = Attachment(buffer=message.to_bytes())
                sub_attach.file_name = '%s.msg' % subject

                self.message.attachments.append(sub_attach)

    def grab_objs(self):
        return [self.message, self.subject, self.received_time]

    def save(self, file_path):
        if file_path and exists(dirname(file_path)):
            self.message.save(file_path)


class Exchange(Account):
    __auto_renew = None
    __auto_renew_thread = None

    """
    Lets make this a child class of Exchangelib's Account so that we can add renew_session()
    """

    def __init__(self, primary_smtp_address, fullname=None, access_type=None, autodiscover=False, credentials=None,
                 config=None, locale=None, default_timezone=None, auto_renew=False):
        Account.__init__(self, primary_smtp_address=primary_smtp_address, fullname=fullname, access_type=access_type,
                         autodiscover=autodiscover, credentials=credentials, config=config, locale=locale,
                         default_timezone=default_timezone)

        if auto_renew:
            self.start_auto_renew()

    def __auto_renew_session(self):
        from time import sleep
        self.__auto_renew = True
        counter = 0

        while self.__auto_renew:
            if counter > 70:
                counter = 0
                self.renew_session()

            sleep(1)
            counter += 1

    def start_auto_renew(self):
        if not self.__auto_renew_thread or not self.__auto_renew_thread.is_alive():
            from threading import Thread
            self.__auto_renew_thread = Thread(target=self.__auto_renew_session)
            self.__auto_renew_thread.daemon = True
            self.__auto_renew_thread.start()

    def stop_auto_renew(self):
        self.__auto_renew = False

        if self.__auto_renew_thread and self.__auto_renew_thread.is_alive():
            self.__auto_renew_thread.join()

        self.__auto_renew_thread = None

    def renew_session(self):
        """
        Renew e-mail session by retiring old session and starting new session. Great way to keep the connection live
        """

        session = self.protocol.get_session()

        if session:
            self.protocol.retire_session(session)

    __del__ = stop_auto_renew



from ..filehandler import FileHandler
from csv import reader
from portalocker import Lock
from pandas import DataFrame


class CSVReader(FileHandler):
    def __init__(self, file_path, dialect=None, **fmtparams):
        super().__init__(file_path=file_path, dialect=dialect, **fmtparams)

    def parse(self):
        if self.streams:
            for handler, buffer in self.streams:
                self.__parse(handler=handler, buffer=buffer)
        else:
            return self.__parse()

    def __parse(self, handler=None, buffer=None):
        data = list()
        row_num = 0
        header = None

        with Lock(filename=self.file_path, mode='r') as read_obj:
            csv_reader = reader(read_obj, self.dialect, **self.kwargs)

            for line in csv_reader:
                data.append(line)

                if not header:
                    header = converted_row

                if handler and buffer <= len(data):
                    df = self.__to_df(data)
                    handler(df, row_num - len(df) + 1, row_num)
                    data.clear()
                    data.append(header)

                row_num += 1

        df = self.__to_df(data)

        if not df.empty and handler:
            row_num -= 1
            handler(df, row_num - len(df) + 1, row_num)
        elif not handler:
            return df

    @staticmethod
    def __to_df(data):
        if data:
            df = DataFrame(data)
            new_header = df.iloc[0]
            df = df[1:]
            df.columns = new_header
            return df
        else:
            return DataFrame()



from ..filehandler import FileHandler
from json import loads
from portalocker import Lock
from pandas import DataFrame


class JSONReader(FileHandler):
    def __init__(self, file_path):
        super().__init__(file_path=file_path)

    def parse(self):
        if self.streams:
            for handler, buffer in self.streams:
                self.__parse(handler=handler, buffer=buffer)
        else:
            return self.__parse()

    def __parse(self, handler=None, buffer=None):
        data = list()
        row_num = 0
        header = None

        with Lock(filename=self.file_path, mode='r') as read_obj:
            for line in read_obj:
                data.append(loads(line))

                if not header:
                    header = converted_row

                if handler and buffer <= len(data):
                    handler(data, row_num - len(data) + 1, row_num)
                    data.clear()
                    data.append(header)

                row_num += 1

        if data and handler:
            row_num -= 1
            handler(data, row_num - len(data) + 1, row_num)
        elif not handler:
            return data

    @staticmethod
    def __to_df(data):
        if data:
            df = DataFrame(data)
            new_header = df.iloc[0]
            df = df[1:]
            df.columns = new_header
            return df
        else:
            return DataFrame()
